#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <stdio.h>
#define __USE_ISOC99 1
#include <math.h>
#include "fov.h"

/*
+---++---++---++---+
|	 ||	 ||	 ||	 |
|	 ||	 ||	 ||	 |
|	 ||	 ||	 ||	 |
+---++---++---++---+	2
+---++---++---+#####
|	 ||	 ||	 |#####
|	 ||	 ||	 |#####
|	 ||	 ||	 |#####
+---++---++---+#####X 1	<-- y
+---++---++---++---+
|	 ||	 ||	 ||	 |
| @ ||	 ||	 ||	 |		 <-- srcy centre	 -> dy = 0.5 = y - 0.5
|	 ||	 ||	 ||	 |
+---++---++---++---+	0
0		1		2		3		4
	^						^
	|						|
 srcx					x			 -> dx = 3.5 = x + 0.5
centre

Slope from @ to X.

+---++---++---++---+
|	 ||	 ||	 ||	 |
|	 ||	 ||	 ||	 |
|	 ||	 ||	 ||	 |
+---++---++---++---+ 2
+---++---++---++---+
|	 ||	 ||	 ||	 |
|	 ||	 ||	 ||	 |
|	 ||	 ||	 ||	 |
+---++---++---+X---+ 1	 <-- y
+---++---++---+#####
|	 ||	 ||	 |#####
| @ ||	 ||	 |#####		 <-- srcy centre	 -> dy = 0.5 = y - 0.5
|	 ||	 ||	 |#####
+---++---++---+##### 0
0		1		2		3
	^						^
	|						|
 srcx					x			 -> dx = 2.5 = x - 0.5
centre

Slope from @ to X
*/


/* Types ---------------------------------------------------------- */

typedef enum {
	FOV_LIMIT_CIRCLE_PRECALCULATE,
	FOV_LIMIT_SQUARE,
	FOV_LIMIT_CIRCLE,
	FOV_LIMIT_OCTAGON
} fov_limit_t;

typedef enum {
	FOV_CORNER_NOPEEK,
	FOV_CORNER_PEEK
} fov_corner_peek_t;

typedef enum {
	FOV_OPAQUE_APPLY,
	FOV_OPAQUE_NOAPPLY
} fov_opaque_apply_t;


/* Data ----------------------------------------------------------- */

static unsigned **heights = NULL;
static size_t numheights = 0;
fov_limit_t option_method = FOV_LIMIT_CIRCLE_PRECALCULATE;
fov_corner_peek_t option_peek = FOV_CORNER_NOPEEK;
fov_opaque_apply_t option_opaqueapply = FOV_OPAQUE_APPLY;


/* Options -------------------------------------------------------- */

void fov_set_limit_circular() { option_method = FOV_LIMIT_CIRCLE; }
void fov_set_limit_circular_precalculated() { option_method = FOV_LIMIT_CIRCLE_PRECALCULATE; }
void fov_set_limit_octagonal() { option_method = FOV_LIMIT_OCTAGON; }
void fov_set_limit_square() { option_method = FOV_LIMIT_SQUARE; }

void fov_set_nopeek() { option_peek = FOV_CORNER_NOPEEK; }
void fov_set_peek() { option_peek = FOV_CORNER_PEEK; }

void fov_set_opaque_apply() { option_opaqueapply = FOV_OPAQUE_APPLY; }
void fov_set_opaque_noapply() { option_opaqueapply = FOV_OPAQUE_NOAPPLY; }


/* Circular FOV --------------------------------------------------- */

unsigned *precalculate_heights(unsigned maxdist) {
	unsigned i;
	unsigned *result = (unsigned *)malloc((maxdist+2)*sizeof(unsigned));
	for (i = 0; i <= maxdist; ++i) {
		result[i] = (unsigned)sqrt(maxdist*maxdist - i*i);
	}
	result[maxdist+1] = 0;
	return result;
}

unsigned height(unsigned x, unsigned maxdist) {
	unsigned **newheights;

	if (heights == NULL || maxdist > numheights) {
		newheights = (unsigned **)calloc(maxdist, sizeof(unsigned*));
		memcpy(heights, newheights, numheights);
		free(heights);
		heights = newheights;
		numheights = maxdist;
	}

	if (heights[maxdist-1] == NULL) {
		heights[maxdist-1] = precalculate_heights(maxdist);
	}
	return heights[maxdist-1][x];
}

void fov_free() {
	unsigned i;
	for (i = 0; i < numheights; ++i) {
		free(heights[i]);
		heights[i] = NULL;
	}
	free(heights);
	heights = NULL;
}


/* Data ----------------------------------------------------------- */

typedef struct {
	void *map;
	void *src;
	int srcx, srcy;
		int (*opaque)(void *map, int x, int y);
	void (*apply)(void *map, int x, int y, int dx, int dy, void *src);
	unsigned radius;
	bool corner_peek;
	bool see_walls;
} fov_static_data_type;

/* Slope ---------------------------------------------------------- */

double fov_slope(double dx, double dy)
{
	if (dx != 0.0)
		return dy/dx;
	else
		return 0.0;
}

/* Octants -------------------------------------------------------- */

#define FOV_DEFINE_OCTANT(signx, signy, rx, ry, nx, ny, nf)				\
	void fov_octant_##nx##ny##nf(										\
								 fov_static_data_type *data,			\
								 int dx,								\
								 float startSlope,						\
								 float endSlope							\
								 ) {									\
		static int temp;												\
		int x,y,dy,dy0,dy1;												\
		unsigned h;														\
		bool prevBlocked;												\
																		\
		if (dx == 0) {													\
			fov_octant_##nx##ny##nf(data, dx+1, startSlope, endSlope);	\
			return;														\
		}																\
		if (dx > data->radius) return;									\
																		\
		dy0 = 0.5f + ((float)dx)*startSlope;							\
		dy1 = 0.5f + ((float)dx)*endSlope;								\
																		\
		/*fprintf(stderr, "(%2d) = [%2d .. %2d] (%f .. %f)\n", dx, dy0, dy1, 0.5f + ((float)dx)*startSlope, 0.5f + ((float)dx)*endSlope);*/ \
																		\
		x = data->src##rx signx dx;										\
		y = data->src##ry signy dy0;									\
																		\
		prevBlocked = data->opaque(data->map, rx, ry);					\
																		\
		if (dy0 == 0) {	/* We do straight lines separately, so they don't get done twice. */ \
			++dy0;														\
		}																\
		if (dy1 == dx) { /* We do diagonal lines separately, so they don't get done twice. */ \
			--dy1;														\
		}																\
		if (dy0 > dy1) { /* Temporary hack to prevent broken lines when dy0 > dy1 (?) */ \
			temp = dy0;													\
			dy0 = dy1;													\
			dy1 = temp;													\
		}																\
																		\
		switch (option_method) {										\
		case FOV_LIMIT_CIRCLE_PRECALCULATE:								\
			h = height(dx, data->radius); break;						\
		case FOV_LIMIT_CIRCLE:											\
			h = sqrt(data->radius*data->radius - dx*dx); break;			\
		case FOV_LIMIT_OCTAGON:											\
			h = (data->radius - dx)<<1; break;							\
		default:														\
			h = data->radius; break;									\
		};																\
		if (h < dy1) dy1 = h;											\
																		\
		for (dy = dy0; dy <= dy1; ++dy) {								\
			y = data->src##ry signy dy;									\
																		\
			if (data->opaque(data->map, rx, ry)) {						\
				if (option_opaqueapply == FOV_OPAQUE_APPLY)				\
					data->apply(data->map, rx, ry, dx, dy, data->src);	\
				if (!prevBlocked) {										\
					float endSlopeNext = fov_slope((float)dx + 0.5f, (float)dy - 0.5f);	\
					fov_octant_##nx##ny##nf(data, dx+1, startSlope, endSlopeNext); \
				}														\
				prevBlocked = true;										\
			} else {													\
				data->apply(data->map, rx, ry, dx, dy, data->src);		\
				if (prevBlocked) {										\
					startSlope = fov_slope((float)dx - 0.5f, (float)dy - 0.5f);	\
				}														\
				prevBlocked = false;									\
			}															\
		}																\
																		\
		if (!prevBlocked) {												\
			fov_octant_##nx##ny##nf(data, dx+1, startSlope, endSlope);	\
		}																\
	}

FOV_DEFINE_OCTANT(+,+,x,y,p,p,n)
FOV_DEFINE_OCTANT(+,+,y,x,p,p,y)
FOV_DEFINE_OCTANT(+,-,x,y,p,m,n)
FOV_DEFINE_OCTANT(+,-,y,x,p,m,y)
FOV_DEFINE_OCTANT(-,+,x,y,m,p,n)
FOV_DEFINE_OCTANT(-,+,y,x,m,p,y)
FOV_DEFINE_OCTANT(-,-,x,y,m,m,n)
FOV_DEFINE_OCTANT(-,-,y,x,m,m,y)


/* Circle --------------------------------------------------------- */

#define HORIZONTAL(signx)									\
	 for (i = 1; i < radius; ++i) {							\
		 x = srcx signx i;									\
		 if (opaque(map, x, srcy)) {						\
			 if (option_opaqueapply == FOV_OPAQUE_APPLY)	\
				 apply(map, x, srcy, signx i, 0, src);		\
			 break;											\
		 }													\
		 apply(map, x, srcy, signx i, 0, src);				\
	 }														\

#define VERTICAL(signy)										\
	 for (i = 1; i < radius; ++i) {							\
		 y = srcy signy i;									\
		 if (opaque(map, srcx, y)) {						\
			 if (option_opaqueapply == FOV_OPAQUE_APPLY)	\
				 apply(map, srcx, y, 0, signy i, src);		\
			 break;											\
		 }													\
		 apply(map, srcx, y, 0, signy i, src);				\
	 }														\

	 
#define DIAGONAL(signx, signy)								\
	 for (i = 1; i < dradius; ++i) {						\
		 x = srcx signx i; y = srcy signy i;				\
		 if (opaque(map, x, y)) {							\
			 if (option_opaqueapply == FOV_OPAQUE_APPLY)	\
				 apply(map, x, y, signx i, signy i, src);	\
			 break;											\
		 }													\
		 apply(map, x, y, signx i, signy i, src);			\
	 }


void fov_circle(void *map, void *src, int srcx, int srcy, unsigned radius,
	int (*opaque)(void *map, int x, int y),
	void (*apply)(void *map, int x, int y, int dx, int dy, void *src))
{
	int x, y, i;
	unsigned dradius;
	fov_static_data_type data;
	data.map = map;
	data.src = src;
	data.srcx = srcx;
	data.srcy = srcy;
	data.opaque = opaque;
	data.apply = apply;
	data.radius = radius;

	if (option_method == FOV_LIMIT_OCTAGON)
		dradius = (radius<<1)/3 + 1;
	else
		dradius = 0.70710678118 * radius + 1;

	HORIZONTAL(+); /* east */
	HORIZONTAL(-); /* west */
	VERTICAL(+); /* south */
	VERTICAL(-); /* north */
	DIAGONAL(+, +); /* south east */
	DIAGONAL(-, +); /* south west */
	DIAGONAL(+, -); /* north east */
	DIAGONAL(-, -); /* north west */

	/*fprintf(stderr, "----------\n");*/

	fov_octant_pmn(&data, 0, 0.0, 1.0);
	fov_octant_ppn(&data, 0, 0.0, 1.0);
	fov_octant_ppy(&data, 0, 0.0, 1.0);
	fov_octant_pmn(&data, 0, 0.0, 1.0);
	fov_octant_pmy(&data, 0, 0.0, 1.0);
	fov_octant_mpn(&data, 0, 0.0, 1.0);
	fov_octant_mpy(&data, 0, 0.0, 1.0);
	fov_octant_mmn(&data, 0, 0.0, 1.0);
	fov_octant_mmy(&data, 0, 0.0, 1.0);
}

void fov_dummy(void *map, void *src, int srcx, int srcy, unsigned radius,
	int (*opaque)(void *map, int x, int y),
	void (*apply)(void *map, int x, int y, int dx, int dy, void *src))
{
}
